#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2020, Authors
# Full license can be found in License.md
# -----------------------------------------------------------------------------

from setuptools import setup


# Run setup.  Setuptools will look for parameters in [metadata] section of
# setup.cfg
setup()
